/*----------------------------------------------------------------------
  File    : ARuleGUI.java
  Contents: association rule induction toolbox
  Author  : Christian Borgelt
  History : 2004.09.29 file created from file DTreeGUI.java
            2007.05.08 adapted to new executor classes
            2007.05.30 mincnt and maxcnt made JSpinners
            2007.06.05 number inputs turned into JFormattedTextFields
            2007.07.07 adapted to new class DialogPanel
            2013.04.22 adapted to type argument of JComboBox
            2013.11.26 adapted to option changes of apriori program
            2014.04.12 adapted to item appearances file option -R
            2014.10.07 selection of apriori, eclat or fpgrowth added
            2014.10.22 terminal tab for external program output added
            2014.10.23 changed from LGPL license to MIT license
            2015.03.30 workaround for Java bug added (command args.)
            2018.11.14 replaced some deprecated functions
----------------------------------------------------------------------*/
package fim;

import java.io.IOException;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JTextField;
import javax.swing.JSpinner;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;

import dialog.TabbedGUI;
import dialog.DialogPanel;
import dialog.FormatPanel;
import dialog.AboutPanel;
import dialog.FormatDialog;
import util.Executor;
import util.CmdExecutor;

/*--------------------------------------------------------------------*/
/** Class for a user interface to association rule mining programs.
 *  @author Christian Borgelt
 *  @since  2004.09.29 */
/*--------------------------------------------------------------------*/
public class ARuleGUI extends TabbedGUI implements ActionListener {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00020010L;
  /** the version string */
  public  static final String VERSION = "2.16 (2018.11.14)";

  /** the index of the Apriori program */
  private static final int APRIORI  = 0;
  /** the index of the Eclat program */
  private static final int ECLAT    = 1;
  /** the index of the FP-growth program */
  private static final int FPGROWTH = 2;

  /** the names of the algorithms/programs */
  private static final String[] algonames = {
    "Apriori", "Eclat", "FP-growth" };
  /** the names of the programs */
  private static final String[] prgnames = {
    "apriori", "eclat", "fpgrowth" };

  /** the names of the additional rule evaluation measures */
  private static final String[] emnames = {
    "none",                                     /* x */
    "rule support",                             /* o */
    "rule confidence",                          /* c */
    "abs. confidence difference to prior",      /* d */
    "lift value (confidence / prior)",          /* l */
    "abs. difference of lift value to 1",       /* a */
    "difference of lift quotient to 1",         /* q */
    "conviction (inv. lift for negated head)",  /* v */
    "abs. difference of conviction to 1",       /* e */
    "diff. of conviction quotient to 1",        /* r */
    "certainty factor (rel. conf. change)",     /* z */
    "normalized chi^2 measure",                 /* n */
    "p-value from chi^2 measure",               /* p */
    "chi^2 measure with Yates' correction",     /* y */
    "p-value from Yates-corrected chi^2",       /* t */
    "information difference to prior",          /* i */
    "p-value from G statistic/info. diff.",     /* g */
    "Fisher's exact test (table probability)",  /* f */
    "Fisher's exact test (chi^2 measure)",      /* h */
    "Fisher's exact test (information gain)",   /* m */
    "Fisher's exact test (support)",            /* s */
  };

  /** the codes of the additional rule evaluation measures */
  private static final char[] emcodes = {
    'x', 'o', 'c', 'd', 'l', 'a', 'q', 'v', 'e', 'r', 'z',
    'n', 'p', 'y', 't', 'i', 'g', 'f', 'h', 'm', 's' };

  /** the names of the transaction sorting modes */
  private static final String[] sortnames = {
    "descending w.r.t. transaction size sum",
    "descending w.r.t. frequency",
    "no sorting",
    "ascending w.r.t. frequency",
    "ascending w.r.t. transaction size sum" };

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /* --- files --- */
  /** the transactions file */
  private JTextField        fn_tra;
  /** the association rule file */
  private JTextField        fn_rul;
  /** the item appearances file */
  private JTextField        fn_app;
  /** the algorithm/program to use */
  private JComboBox<String> algo;
  /** the path to the C programs */
  private File              path;

  /* --- filters --- */
  /** the minimum number of items per rule */
  private JSpinner          mincnt;
  /** the maximum number of items per rule */
  private JSpinner          maxcnt;
  /** the minimum support of a rule */
  private JTextField        minsupp;
  /** the maximum support of a rule */
  private JTextField        maxsupp;
  /** whether to use the original support definition */
  private JCheckBox         orig;
  /** the minimum confidence of a rule */
  private JTextField        minconf;
  /** the additional rule evaluation measure */
  private JComboBox<String> eval;
  /** the threshold for the additional rule evaluation measure */
  private JTextField        thresh;

  /* --- output --- */
  /** the record header for the rule output */
  private JTextField        header;
  /** the item separator for the rule output */
  private JTextField        itemsep;
  /** the implication sign for the rule output */
  private JTextField        implic;
  /** the output format for the rule information */
  private JTextField        outfmt;

  /* --- options --- */
  /** the sorting mode for the transactions */
  private JComboBox<String> sort;
  /** whether to organize the transactions as a prefix tree */
  private JCheckBox         prefix;
  /** whether to use a-posteriori pruning */
  private JCheckBox         post;
  /** how to filter unused items */
  private JTextField        filter;

  /*------------------------------------------------------------------*/
  /** Create an association rule GUI.
   *  <p>The dialog is created as a stand-alone program.
   *  That is, closing or quitting it terminates the program.</p>
   *  @since  2004.09.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public ARuleGUI ()
  { this.init(null, true); }

  /*------------------------------------------------------------------*/
  /** Create an association rule GUI.
   *  @param  owner the component that is to own this dialog
   *  @since  2004.09.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public ARuleGUI (Component owner)
  { this.init(owner, false); }

  /*------------------------------------------------------------------*/
  /** Create the user interface.
   *  <p>Following the recommendations in the Java tutorial, the user
   *  interface is created in the "run" method, which is invoked from
   *  the event queue, in order to avoid problems with threads.</p>
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  {                             /* --- create the user interface */
    DialogPanel tab;            /* current tab of the tabbed pane */

    /* --- create basic GUI --- */
    this.base("Association Rule Induction Tools");
    this.addFormatTab(FormatPanel.TAWGT);

    /* --- Files --- */
    tab = this.addTab("Files");

    tab.addLabel("Transaction file:");
    tab.addButton("Select", DialogPanel.MIDDLE)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.getFileName(ARuleGUI.this.fn_tra); } } );
    tab.addButton("View", DialogPanel.RIGHT)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.showTrActs(ARuleGUI.this.fn_tra); } } );
    this.fn_tra = tab.addFileInput("data"+File.separator+"tracts.txt");

    tab.addLabel("Association rule file:");
    tab.addButton("Select", DialogPanel.MIDDLE)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.getFileName(ARuleGUI.this.fn_rul); } } );
    tab.addButton("View", DialogPanel.RIGHT).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.showRules(ARuleGUI.this.fn_rul); } } );
    this.fn_rul = tab.addFileInput("data"+File.separator+"rules.txt");

    tab.addLabel("Item appearances file:");
    tab.addButton("Select", DialogPanel.MIDDLE)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.getFileName(ARuleGUI.this.fn_app); } } );
    tab.addButton("Edit", DialogPanel.RIGHT)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        ARuleGUI.this.editFile(ARuleGUI.this.fn_app); } } );
    this.fn_app = tab.addFileInput("");
    tab.addHelp("The item appearances file is optional "
               +"and may be missing.\nThis file describes "
               +"in which parts of an association rule\n"
               +"(antecedent/consequent/both/neither) an item "
               +"may appear.");
    tab.addFiller(0);

    tab.addLabel("Algorithm/program:");
    this.algo = tab.addComboBox(ARuleGUI.algonames);
    this.algo.setSelectedIndex(FPGROWTH);

    tab.addButton("Locate Programs...", DialogPanel.RIGHT)
      .addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JFileChooser c = ARuleGUI.this.getFileChooser();
        c.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        c.setDialogTitle("Locate Program...");
        int r = c.showOpenDialog(ARuleGUI.this);
        if (r == JFileChooser.APPROVE_OPTION)
          ARuleGUI.this.path = c.getSelectedFile();
        c.setFileSelectionMode(JFileChooser.FILES_ONLY);
        c.setDialogTitle("Open File...");
      } } );

    /* --- Filters --- */
    tab = this.addTab("Filters");

    tab.addLabel("Minimum number of items:");
    this.mincnt = tab.addSpinner(1, 1, 999999, 1);
    tab.addLabel("Maximum number of items:");
    this.maxcnt = tab.addSpinner(0, 0, 999999, 1);

    tab.addLabel("Minimum support [#/%]:");
    this.minsupp = tab.addNumberInput("10");
    tab.addLabel("Maximum support [#/%]:");
    this.maxsupp = tab.addNumberInput("100");
    tab.addHelp("A negative support means an absolute number (#).");
    tab.addLabel("Minimum confidence [%]:");
    this.minconf = tab.addNumberInput("80");

    tab.addLabel("Original support definition:");
    this.orig = tab.addCheckBox(false);
    tab.addHelp("Use the support of all occurring items as the support "
               +"of an\nassociation rule and not only the support "
               +"of the antecedent.");
    tab.addFiller(4);

    tab.addLabel("Additional evaluation measure:", DialogPanel.RIGHT);
    this.eval = tab.addComboBox(ARuleGUI.emnames);
    tab.addLabel("Threshold [%]:");
    this.thresh = tab.addNumberInput("10");
    tab.addFiller(0);

    /* --- Output --- */
    tab = this.addTab("Output");
    tab.addLabel("Rule header:");
    tab.add(this.header = new JTextField(""), DialogPanel.RIGHT);
    tab.addHelp("String that is printed before each rule.");
    tab.addLabel("Implication sign:");
    tab.add(this.implic = new JTextField(" <- "), DialogPanel.RIGHT);
    tab.addHelp("String to separate rule consequent and antecedent.");
    tab.addLabel("Item separator:");
    tab.add(this.itemsep = new JTextField(" "), DialogPanel.RIGHT);
    tab.addHelp("String to separate the items of the rule antecedent.");
    tab.addLabel("Information output format:", DialogPanel.FILL);
    this.outfmt = tab.addTextInput(" (%b,%a,%h,%Q,%E)",
                                   DialogPanel.FILL);
    tab.addHelp("The information output format is very important for "
               +"the\nparser of the association rule viewer to work "
               +"correctly.\nChange at most the number of significant "
               +"digits and/or\nthe letter E to lowercase if you want "
               +"to use the viewer.\nLikewise, do not change the rule "
               +"header, implication sign\nor item separator "
               +"in this case.");
    tab.addFiller(0);

    /* --- Options --- */
    tab = this.addTab("Options");
    tab.addLabel("Item sorting:", DialogPanel.RIGHT);
    this.sort = tab.addComboBox(ARuleGUI.sortnames);
    this.sort.setSelectedIndex(4);
    tab.addHelp("An item order other than the default "
               +"is usually slower.");

    tab.addLabel("Transaction prefix tree:");
    this.prefix = tab.addCheckBox(true);
    tab.addHelp("Representing the transaction database as a prefix "
               +"tree\nusually speeds up processing "
               +"(only Apriori).");

    tab.addLabel("Filter unused items:");
    this.filter = tab.addTextInput("0.01");
    tab.addHelp("< 0: fraction of removed items to trigger filtering,\n"
               +"> 0: take execution times ratio into account "
               +"(only Apriori).");

    tab.addLabel("A-posteriori pruning:");
    this.post = tab.addCheckBox(false);
    tab.addHelp("Remove infrequent item sets from the data structure\n"
               +"(only Apriori).");
    tab.addFiller(0);

    /* --- Terminal --- */
    this.addTerminalTab("Terminal output of external programs");

    /* --- About --- */
    this.addTab("About", new AboutPanel(
       "Association Rule Induction Tools",
       "A simple user interface for association rule induction.\n\n"
      +"Version " +ARuleGUI.VERSION +"\n"
      +"written by Christian Borgelt\n"
      +"European Centre for Soft Computing\n"
      +"c/ Gonzalo Gutierrez Quiros s/n\n"
      +"33600 Mieres, Asturias, Spain\n"
      +"christian@borgelt.net"));

    /* --- finalize --- */
    this.pack();
    this.selectTab(1);
  }  /* run() */

  /*------------------------------------------------------------------*/
  /** Get the path to the association rule mining programs.
   *  @return the path to the association rule mining programs
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public File getPath ()
  { return this.path; }

  /*------------------------------------------------------------------*/
  /** Set the path to the association rule mining programs.
   *  @param  path the path to the association rule mining programs
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setPath (File path)
  { this.path = path; }

  /*------------------------------------------------------------------*/
  /** Set the transactions file.
   *  @param  file the transactions file to set
   *  @since  2014.09.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrActsFile (File file)
  { this.fn_tra.setText(file.getPath()); }

  /*------------------------------------------------------------------*/
  /** Set the association rule output file.
   *  @param  file the association rule output file to set
   *  @since  2014.09.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setOutputFile (File file)
  { this.fn_rul.setText(file.getPath()); }

  /*------------------------------------------------------------------*/
  /** Set the item appearances file.
   *  @param  file the item appearances file to set
   *  @since  2014.09.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setItemAppFile (File file)
  { this.fn_app.setText(file.getPath()); }

  /*------------------------------------------------------------------*/
  /** Show association rules.
   *  @param  txt the text field containing the file name
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showRules (JTextField txt)
  { this.showRules(new File(txt.getText())); }

  /*------------------------------------------------------------------*/
  /** Show a set of association rules.
   *  @param  file the file to load the association rules from
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showRules (File file)
  {                             /* --- show a set of ass. rules */
    ARuleView view = new ARuleView(this, ARuleView.FILE_ITEMS);
    if (!view.loadRules(file)) return;
    view.setVisible(true); view.toFront();
  }  /* showRules() */

  /*------------------------------------------------------------------*/
  /** Show a set of transactions.
   *  @param  txt the text field containing the file name
   *  @since  2007.06.06 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showTrActs (JTextField txt)
  { this.showTrActs(new File(txt.getText())); }

  /*------------------------------------------------------------------*/
  /** Show a set of transactions.
   *  @param  file the file to load the transactions from
   *  @since  2007.06.06 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showTrActs (File file)
  {                             /* --- show a set of ass. rules */
    TrActView    view;          /* viewer for the transactions */
    FormatDialog fmtdlg;        /* format dialog of the viewer */

    view = new TrActView(this, TrActView.FILE_ITEMS);
    if (this.format != null) {  /* if there is a format tab */
      fmtdlg = view.getFormatDlg();
      fmtdlg.setMode   (this.format.getMode());
      fmtdlg.setRecSeps(this.format.getRecSeps());
      fmtdlg.setFldSeps(this.format.getFldSeps());
      fmtdlg.setBlanks (this.format.getBlanks());
      fmtdlg.setComment(this.format.getComment());
    }                           /* transfer characters */
    if (!view.loadTrActs(file)) return;
    view.setVisible(true); view.toFront();
  }  /* showTrActs() */

  /*------------------------------------------------------------------*/
  /** Get an executor for the association rule induction program.
   *  @return the executor for the association rule program
   *  @since  2007.03.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Executor createExecutor (int i)
  {                             /* --- get the command to execute */
    int      n = 0, k;          /* number of arguments, buffer */
    int      algo;              /* index of algorithm/program */
    String[] cmd;               /* command and arguments */
    File     pp;                /* the path to the programs */
    String   s;                 /* buffer for arguments */

    cmd    = new String[64];    /* create a command array */
    pp     = this.getPath();    /* and get the program path */
    algo   = this.algo.getSelectedIndex();
    cmd[0] = ((pp != null) ? pp +File.separator : "")
           + ARuleGUI.prgnames[algo];
    n = this.format.addFormatArgs(cmd, FormatPanel.TAWGT);
    cmd[n++] = "-m" +this.mincnt.getValue();
    k = ((Integer)this.maxcnt.getValue()).intValue();
    if (k > 0) cmd[n++] = "-n" +k;
    if (this.orig.isSelected()) cmd[n++] = "-o";
    s = this.minsupp.getText().trim(); /* minimum support */
    if (s.length() > 0) cmd[n++] = "-s" +s;
    s = this.maxsupp.getText().trim(); /* maximum support */
    if (s.length() > 0) cmd[n++] = "-S" +s;
    s = this.minconf.getText().trim(); /* minimum confidence */
    if (s.length() > 0) cmd[n++] = "-c" +s;
    cmd[n++] = "-gtr";          /* scanable form for items */
    cmd[n++] = "-h"; cmd[n++] = this.header.getText();
    if (cmd[n-1].length() <= 0) cmd[n-1] = "\"\"";
    cmd[n++] = "-I"; cmd[n++] = this.implic.getText();
    if (cmd[n-1].length() <= 0) cmd[n-1] = "\"\"";
    cmd[n++] = "-k"; cmd[n++] = this.itemsep.getText();
    if (cmd[n-1].length() <= 0) cmd[n-1] = "\"\"";
    cmd[n++] = "-v"; cmd[n++] = this.outfmt.getText();
    if (cmd[n-1].length() <= 0) cmd[n-1] = "\"\"";
    k = this.eval.getSelectedIndex();
    if (k > 0) {                /* rule evaluation measure */
      cmd[n++] = "-e" +ARuleGUI.emcodes[k];
      cmd[n++] = "-d" +this.thresh.getText();
    }                           /* threshold for evaluation measure */
    k = this.sort.getSelectedIndex() -2;
    if (k != 2) cmd[n++] = "-q" +k;
    if (algo == APRIORI) {      /* if apriori algorithm/program */
      cmd[n++] = "-u" +this.filter.getText();
      if (!this.prefix.isSelected()) cmd[n++] = "-T";
      if ( this.post.isSelected())   cmd[n++] = "-y";
    }                           /* add apriori-specific options */
    s = fn_app.getText();       /* add the file arguments */
    if (s.length() > 0) cmd[n++] = "-R" +s;
    cmd[n++] = fn_tra.getText();
    cmd[n++] = fn_rul.getText();
    return new CmdExecutor(TabbedGUI.shrinkCmd(cmd, n), this);
  }  /* createExecutor() */

  /*------------------------------------------------------------------*/
  /** Get the result message on successful termination.
   *  @return the result message
   *  @since  2007.02.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected String getResultMsg ()
  {                             /* --- get result message */
    int    i, k;                /* index variables */
    String msg;                 /* buffer for message */
    float  t;                   /* execution time */

    msg = ((CmdExecutor)this.executor).getLastErrorLine();
    i = msg.lastIndexOf("[", msg.indexOf("rule(s)]"));
    k = msg.indexOf(']', i);    /* find rules report */
    t = (System.currentTimeMillis() -this.start) / 1000.0F;
    return "Found " +msg.substring(i+1, k) +".\n"
          +"Total search time: " +t +"s.\n"
          +"(See terminal for more information.)";
  }  /* getResultMessage() */

  /*------------------------------------------------------------------*/
  /** Load a configuration file and set the input fields.
   *  @param  file the file to load the configuration from
   *  @since  2007.07.08 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void loadConfig (File file)
  {                             /* --- load configuration file */
    if (file == null) {         /* if no file name is given */
      JFileChooser c = this.getFileChooser();
      c.setDialogTitle("Load Configuration...");
      int r = c.showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return;
      file = c.getSelectedFile();
    }                           /* let the user choose a file */
    try {                       /* open the configuration file */
      FileReader reader = new FileReader(file);
      /* --- Format --- */
      this.format.loadConfig(reader);
      /* --- text parameters --- */
      /* --- Files --- */
      this.fn_tra.setText (this.readLine(reader));
      this.fn_rul.setText (this.readLine(reader));
      this.fn_app.setText (this.readLine(reader));
      /* --- Filters --- */
      this.minsupp.setText(this.readLine(reader));
      this.maxsupp.setText(this.readLine(reader));
      this.minconf.setText(this.readLine(reader));
      this.thresh.setText (this.readLine(reader));
      this.filter.setText (this.readLine(reader));
      /* --- Output --- */
      this.header.setText (this.readLine(reader));
      this.implic.setText (this.readLine(reader));
      this.itemsep.setText(this.readLine(reader));
      this.outfmt.setText (this.readLine(reader));
      /* --- flags & indices --- */
      /* --- Filters --- */
      this.mincnt.setValue(Integer.valueOf(this.readInt(reader)));
      this.maxcnt.setValue(Integer.valueOf(this.readInt(reader)));
      this.eval.setSelectedIndex(this.readInt(reader));
      this.sort.setSelectedIndex(this.readInt(reader));
      /* --- Options --- */
      this.orig.setSelected     (this.readInt(reader) != 0);
      this.prefix.setSelected   (this.readInt(reader) != 0);
      this.post.setSelected     (this.readInt(reader) != 0);
      reader.close(); }         /* read the configuration values */
    catch (IOException e) {     /* and close the input file */
      JOptionPane.showMessageDialog(this,
        "Error reading configuration file:\n" +e.getMessage(),
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* check for successful reading */
    this.status.setText("configuration loaded: " +file.getName());
  }  /* loadConfig() */

  /*------------------------------------------------------------------*/
  /** Save a configuration file
   *  @param  file the file to save the current configuration to
   *  @since  2007.07.08 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void saveConfig (File file)
  {                             /* --- save configuration file */
    if (file == null) {         /* if no file name is given, */
      JFileChooser c = this.getFileChooser();
      c.setDialogTitle("Save Configuration...");
      int r = c.showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return;
      file = c.getSelectedFile();
    }                           /* let the user choose a file */
    try {                       /* open the configuration file */
      FileWriter writer = new FileWriter(file);
      /* --- Format --- */
      this.format.saveConfig(writer);
      /* --- text parameters --- */
      /* --- Files --- */
      writer.write(this.fn_tra.getText());  writer.write('\n');
      writer.write(this.fn_rul.getText());  writer.write('\n');
      writer.write(this.fn_app.getText());  writer.write('\n');
      /* --- Filters --- */
      writer.write(this.minsupp.getText()); writer.write('\n');
      writer.write(this.maxsupp.getText()); writer.write('\n');
      writer.write(this.minconf.getText()); writer.write('\n');
      writer.write(this.thresh.getText());  writer.write('\n');
      writer.write(this.filter.getText());  writer.write('\n');
      /* --- Output --- */
      writer.write(this.header.getText());  writer.write('\n');
      writer.write(this.implic.getText());  writer.write('\n');
      writer.write(this.itemsep.getText()); writer.write('\n');
      writer.write(this.outfmt.getText());  writer.write('\n');
      /* --- flags & indices --- */
      /* --- Filters --- */
      writer.write(((Integer)this.mincnt.getValue()).intValue() +",");
      writer.write(((Integer)this.maxcnt.getValue()).intValue() +",");
      writer.write(this.eval.getSelectedIndex() +",");
      writer.write(this.sort.getSelectedIndex() +",");
      /* --- Options --- */
      writer.write(this.orig.isSelected()   ? "1," : "0,");
      writer.write(this.prefix.isSelected() ? "1," : "0,");
      writer.write(this.post.isSelected()   ? "1," : "0,");
      writer.write('\n');       /* write the configuration values */
      writer.close(); }         /* and close the output file */
    catch (IOException e) {     /* check for successful writing */
      JOptionPane.showMessageDialog(this,
        "Error writing configuration file:\n" +e.getMessage(),
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* show a status message */
    this.status.setText("configuration saved: " +file.getName());
  }  /* saveConfig() */

  /*------------------------------------------------------------------*/
  /** Main function for command line invocation.
   *  @param  args the command line arguments
   *  @since  2004.09.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function */
    ARuleGUI gui = new ARuleGUI();
    if (args.length > 0)        /* load configuration if necessary */
      gui.loadConfig(new File(args[0]));
    gui.setVisible(true);       /* show the user interface */
  }  /* main() */

}  /* class ARuleGUI */
